from copy import deepcopy
from math import isclose
from PyCompGeomAlgorithms.core import Point
from PyCompGeomAlgorithms.graham import GrahamStepsTable, GrahamStepsTableRow
from AlgoGrade.graham import GrahamGrader, GrahamTask, GrahamAnswers
from AlgoGrade.adapters import pycga_to_pydantic
from AlgoGrade.core import Scoring


points = [
    Point(2, 8),
    Point(5, 6),
    Point(7, 8),
    Point(8, 11),
    Point(7, 5),
    Point(10, 7),
    Point(11, 5),
    Point(8, 2),
    Point(1, 3),
    Point(5, 2),
]
givens = (points,)
task_class = GrahamTask
scorings = [
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.15, fine=0.15),
    Scoring(max_grade=0.15, fine=0.15),
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.6, fine=0.3, repeat_fine=0.6),
    Scoring(max_grade=0.1, fine=0.25)
]
correct_pycga_answers = task_class.solve_as_pycga_list(givens)
correct_pydantic_answers = task_class.solve_as_pydantic_list(givens)
correct_answers_wrapper = task_class.solve_as_answers_wrapper(givens)


def test_graham_grader_all_correct():
    centroid = Point(4.6667, 7.3333)
    ordered = [
        Point(8, 2),
        Point(7, 5),
        Point(11, 5),
        Point(10, 7),
        Point(7, 8),
        Point(8, 11),
        Point(2, 8),
        Point(1, 3),
        Point(5, 2),
        Point(5, 6)
    ]
    origin = Point(8, 2)
    steps_table = GrahamStepsTable(ordered)
    steps_table.extend([
        GrahamStepsTableRow((ordered[0], ordered[1], ordered[2]), False),
        GrahamStepsTableRow((ordered[0], ordered[2], ordered[3]), True),
        GrahamStepsTableRow((ordered[2], ordered[3], ordered[4]), True),
        GrahamStepsTableRow((ordered[3], ordered[4], ordered[5]), False),
        GrahamStepsTableRow((ordered[2], ordered[3], ordered[5]), False),
        GrahamStepsTableRow((ordered[0], ordered[2], ordered[5]), True),
        GrahamStepsTableRow((ordered[2], ordered[5], ordered[6]), True),
        GrahamStepsTableRow((ordered[5], ordered[6], ordered[7]), True),
        GrahamStepsTableRow((ordered[6], ordered[7], ordered[8]), True),
        GrahamStepsTableRow((ordered[7], ordered[8], ordered[9]), True),
        GrahamStepsTableRow((ordered[8], ordered[9], ordered[0]), False),
        GrahamStepsTableRow((ordered[7], ordered[8], ordered[0]), True)
    ])
    triples = [row.point_triple for row in steps_table]
    are_angles_less_than_pi = [row.is_angle_less_than_pi for row in steps_table]

    pycga_answers = [
        centroid,
        ordered,
        origin,
        triples,
        are_angles_less_than_pi,
        steps_table,
        steps_table,
        steps_table
    ]
    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 2)

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 2)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 2)


def test_graham_grader_incorrect_centroid():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.75)


def test_graham_grader_incorrect_ordered_points():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[1] = [
        Point(2000, 2000),
        Point(5000, 5000),
        Point(11, 5),
        Point(10, 7),
        Point(7, 8),
        Point(8, 11),
        Point(2, 8),
        Point(3000, 3000),
        Point(5, 2),
        Point(5, 6)
    ]

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.75)


def test_graham_grader_incorrect_origin():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[2] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.75)


def test_graham_grader_incorrect_triples():
    pycga_answers = deepcopy(correct_pycga_answers)
    ordered = pycga_answers[1]
    pycga_answers[3] = [
        (ordered[0], ordered[0], ordered[0]),
        (ordered[0], ordered[0], ordered[0]),
        (ordered[2], ordered[3], ordered[4]),
        (ordered[3], ordered[4], ordered[5]),
        (ordered[2], ordered[3], ordered[5]),
        (ordered[0], ordered[2], ordered[5]),
        (ordered[2], ordered[5], ordered[6]),
        (ordered[5], ordered[6], ordered[7]),
        (ordered[6], ordered[7], ordered[8]),
        (ordered[7], ordered[8], ordered[9]),
        (ordered[8], ordered[9], ordered[0]),
        (ordered[7], ordered[8], ordered[0])
    ]

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.85)

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.85)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.85)


def test_graham_grader_incorrect_are_angles_less_than_pi():
    pycga_answers = deepcopy(correct_pycga_answers)
    pycga_answers[4] = [False for _ in pycga_answers[4]]

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.85)

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.85)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.85)


def test_graham_grader_incorrect_rows_with_angles_less_than_pi():
    pycga_answers = deepcopy(correct_pycga_answers)
    ordered = pycga_answers[1]
    incorrect_point = Point(1000, 1000)

    pycga_answers[5][7] = GrahamStepsTableRow((ordered[5], ordered[6], incorrect_point), True)
    pycga_answers[5][8] = GrahamStepsTableRow((ordered[6], ordered[7], incorrect_point), True)
    
    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.75)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.75)


def test_graham_grader_incorrect_rows_with_angles_not_less_than_pi_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    ordered = pycga_answers[1]
    incorrect_point = Point(1000, 1000)

    pycga_answers[5][1] = GrahamStepsTableRow((ordered[0], ordered[2], incorrect_point), True)
    
    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.45) # also triggers "rows with angles < pi" grading

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.45)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.45)


def test_graham_grader_incorrect_rows_with_angles_not_less_than_pi_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    incorrect_point = Point(1000, 1000)

    pycga_answers[5][1] = GrahamStepsTableRow((incorrect_point, incorrect_point, incorrect_point), True)
    pycga_answers[5][5] = GrahamStepsTableRow((incorrect_point, incorrect_point, incorrect_point), True)
    
    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.15) # also triggers "rows with angles < pi" grading

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.15)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.15)


def test_graham_grader_incorrect_finalization():
    pycga_answers = deepcopy(correct_pycga_answers)
    ordered = pycga_answers[1]
    pycga_answers[6][7] = GrahamStepsTableRow((ordered[5], ordered[0], ordered[7]), True)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = GrahamAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = GrahamGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1.5) # also triggers "rows with angles < pi" grading

    total_grade, answer_grades = GrahamGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1.5)

    total_grade, answer_grades = GrahamGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1.5)





from __future__ import annotations
from functools import partial
from typing import ClassVar, Optional, Union
from PyCompGeomAlgorithms.core import BinTree, BinTreeNode, ThreadedBinTree, ThreadedBinTreeNode
from PyCompGeomAlgorithms.dynamic_hull import upper_dynamic_hull, DynamicHullNode, DynamicHullTree, SubhullNode, SubhullThreadedBinTree, PathDirection
from .adapters import pycga_to_pydantic, pydantic_to_pycga, PointPydanticAdapter, BinTreeNodePydanticAdapter, BinTreePydanticAdapter, ThreadedBinTreeNodePydanticAdapter, ThreadedBinTreePydanticAdapter
from .core import Task, Grader, Answers, Mistake
from .parsers import PointListAndTargetPointGivenJSONParser


class DynamicHullGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [
            cls.grade_iterable,
            partial(
                cls.grade_bin_tree,
                grade_item_method=lambda a, c, gp: cls.grade_iterable(
                    (a.left_supporting, a.right_supporting),
                    (c.left_supporting, c.right_supporting),
                    gp
                )
            ),
            cls.grade_omitted_points,
            partial(
                cls.grade_bin_tree,
                grade_item_method=lambda a, c, gp: cls.grade_iterable(
                    [n.point for n in a.subhull.traverse_inorder()],
                    [n.point for n in c.subhull.traverse_inorder()],
                    gp
                )
            ),
            partial(
                cls.grade_bin_tree,
                grade_item_method=lambda a, c, gp: cls.grade_default(a.point, c.point, gp)
            ),
            partial(
                cls.grade_bin_tree,
                grade_item_method=lambda a, c, gp: cls.grade_iterable(
                    a.optimized_subhull.traverse_inorder(),
                    c.optimized_subhull.traverse_inorder(),
                    gp
                )
            ),
            cls.grade_iterable,
            partial(cls.grade_iterable, grade_item_method=[cls.grade_bin_tree, cls.grade_iterable])
        ]
    
    @classmethod
    def grade_omitted_points(cls, answer, correct_answer, scorings):
        def grade_item_method(a, c, gp):
            if not a.is_leaf:
                subhull = [node.point for node in a.subhull.traverse_inorder()]
                left_subhull = [node.point for node in a.left.subhull.traverse_inorder()]
                right_subhull = [node.point for node in a.right.subhull.traverse_inorder()]
                
                try:
                    left_omitted_points = set([point for point in left_subhull[left_subhull.index(a.left_supporting)+1:]])
                    right_omitted_points = set([point for point in right_subhull[:right_subhull.index(a.right_supporting)]])

                    subhull = set(subhull)
                    if subhull & (left_omitted_points | right_omitted_points):
                        return [Mistake(gp)]
                except (IndexError, ValueError):
                    return [Mistake(gp)]

            return []
        
        return cls.grade_bin_tree(answer, correct_answer, scorings, grade_item_method)


class DynamicHullNodePydanticAdapter(BinTreeNodePydanticAdapter):
    regular_class: ClassVar[type] = DynamicHullNode
    data: PointPydanticAdapter
    left: Optional[DynamicHullNodePydanticAdapter] = None
    right: Optional[DynamicHullNodePydanticAdapter] = None
    subhull_points: list[PointPydanticAdapter]
    optimized_subhull_points: Optional[list[PointPydanticAdapter]] = None
    left_supporting_index: int = 0
    left_supporting: PointPydanticAdapter
    right_supporting: PointPydanticAdapter

    @classmethod
    def from_regular_object(cls, obj: DynamicHullNode, **kwargs):
        return super().from_regular_object(
            obj,
            subhull_points=pycga_to_pydantic([node.point for node in obj.subhull.traverse_inorder()]),
            optimized_subhull_points=pycga_to_pydantic([node.point for node in obj.optimized_subhull.traverse_inorder()]),
            left_supporting_index=obj.left_supporting_index,
            left_supporting=pycga_to_pydantic(obj.left_supporting),
            right_supporting=pycga_to_pydantic(obj.right_supporting),
            **kwargs
        )


class DynamicHullTreePydanticAdapter(BinTreePydanticAdapter):
    regular_class: ClassVar[type] = DynamicHullTree
    root: DynamicHullNodePydanticAdapter

    @classmethod
    def from_regular_object(cls, obj: DynamicHullTree, **kwargs):
        return super().from_regular_object(obj, **kwargs)


class SubhullNodePydanticAdapter(ThreadedBinTreeNodePydanticAdapter):
    regular_class: ClassVar[type] = SubhullNode
    data: PointPydanticAdapter
    left: Optional[SubhullNodePydanticAdapter] = None
    right: Optional[SubhullNodePydanticAdapter] = None
    prev: Optional[Union[SubhullNodePydanticAdapter, int]] = None
    next: Optional[Union[SubhullNodePydanticAdapter, int]] = None

    @classmethod
    def from_regular_object(cls, obj: SubhullNode, **kwargs):
        return super().from_regular_object(obj, **kwargs)


class SubhullThreadedBinTreePydanticAdapter(ThreadedBinTreePydanticAdapter):
    regular_class: ClassVar[type] = SubhullThreadedBinTree
    root: SubhullNodePydanticAdapter

    @classmethod
    def from_regular_object(cls, obj: SubhullThreadedBinTree, **kwargs):
        return super().from_regular_object(obj, **kwargs)


class DynamicHullAnswers(Answers):
    leaves: list[DynamicHullNodePydanticAdapter]
    tree: DynamicHullTreePydanticAdapter
    optimized_tree: DynamicHullTreePydanticAdapter
    path: list[PathDirection]
    modified_tree: DynamicHullTreePydanticAdapter
    hull: list[PointPydanticAdapter]

    @classmethod
    def from_iterable(cls, iterable):
        leaves, tree, _, _, _, optimized_tree, path, (modified_tree, hull), *rest = iterable
        return cls(
            leaves=leaves, tree=tree, optimized_tree=optimized_tree,
            path=path, modified_tree=modified_tree, hull=hull
        )
    
    def to_pydantic_list(self):
        return [
            self.leaves, self.tree, self.tree, self.tree, self.tree,
            self.optimized_tree, self.path, (self.modified_tree, self.hull)
        ]


class DynamicHullTask(Task):
    algorithm = upper_dynamic_hull
    grader_class = DynamicHullGrader
    answers_class = DynamicHullAnswers
    given_parser_class = PointListAndTargetPointGivenJSONParser



from AlgoGrade.adapters import PointPydanticAdapter, pydantic_to_pycga
from AlgoGrade.graham import GrahamAnswers, GrahamStepsTableRowPydanticAdapter, GrahamStepsTablePydanticAdapter


def test_graham_answers():
    point = PointPydanticAdapter(coords=(1, 1))
    centroid = point
    ordered_points = [point]
    origin = point
    point_triples = [(point, point, point)]
    are_angles_less_than_pi = [True]
    steps_table = GrahamStepsTablePydanticAdapter(
        ordered_points=[point],
        rows=[GrahamStepsTableRowPydanticAdapter(point_triple=(point, point, point), is_angle_less_than_pi=True)]
    )

    answers_model = GrahamAnswers(
        centroid=centroid, ordered_points=ordered_points, origin=origin, point_triples=point_triples,
        are_angles_less_than_pi=are_angles_less_than_pi, steps_table=steps_table
    )
    answers_list = [
        centroid, ordered_points, origin, point_triples,
        are_angles_less_than_pi, steps_table, steps_table, steps_table
    ]

    assert answers_model.to_pydantic_list() == answers_list
    assert answers_model.to_pycga_list() == pydantic_to_pycga(answers_list)
    assert GrahamAnswers.from_iterable(answers_list) == answers_model
    assert GrahamAnswers(**answers_model.model_dump()) == answers_model



from PyCompGeomAlgorithms.dynamic_hull import PathDirection
from AlgoGrade.adapters import PointPydanticAdapter, pydantic_to_pycga
from AlgoGrade.dynamic_hull import DynamicHullAnswers, DynamicHullNodePydanticAdapter, DynamicHullTreePydanticAdapter


def test_dynamic_hull_answers():
    point = PointPydanticAdapter(coords=(1, 1))
    hull = [point]
    root = DynamicHullNodePydanticAdapter(
        data=point,
        subhull_points=hull,
        left_supporting=point,
        right_supporting=point
    )
    leaves = [root]
    tree = DynamicHullTreePydanticAdapter(root=root)
    optimized_tree = tree
    path = [PathDirection.right]
    modified_tree = tree

    answers_model = DynamicHullAnswers(
        leaves=leaves, tree=tree, optimized_tree=optimized_tree,
        path=path, modified_tree=modified_tree, hull=hull
    )
    answers_list = [leaves, tree, tree, tree, tree, optimized_tree, path, (modified_tree, hull)]

    assert answers_model.to_pydantic_list() == answers_list
    assert answers_model.to_pycga_list() == pydantic_to_pycga(answers_list)
    assert DynamicHullAnswers.from_iterable(answers_list) == answers_model
    assert DynamicHullAnswers(**answers_model.model_dump()) == answers_model


from pytest import fixture
from PyCompGeomAlgorithms.core import Point
from PyCompGeomAlgorithms.graham import GrahamStepsTableRow, GrahamStepsTable
from AlgoGrade.adapters import PointPydanticAdapter
from AlgoGrade.graham import GrahamStepsTableRowPydanticAdapter, GrahamStepsTablePydanticAdapter


@fixture
def table_row_adapter():
    return GrahamStepsTableRowPydanticAdapter(
        point_triple=(
            PointPydanticAdapter(coords=(1, 1)),
            PointPydanticAdapter(coords=(2, 2)),
            PointPydanticAdapter(coords=(3, 3))
        ),
        is_angle_less_than_pi=True
    )


@fixture
def table_adapter(table_row_adapter):
    return GrahamStepsTablePydanticAdapter(
        ordered_points=[PointPydanticAdapter(coords=(1, 1))],
        rows=[table_row_adapter]
    )


def test_steps_table_row_adapter(table_row_adapter):
    regular_object = GrahamStepsTableRow((Point(1, 1), Point(2, 2), Point(3, 3)), True)

    assert table_row_adapter.regular_object() == regular_object
    assert GrahamStepsTableRowPydanticAdapter.from_regular_object(regular_object) == table_row_adapter


def test_steps_table_row_adapter_serializaion(table_row_adapter):
    serialized_row = table_row_adapter.model_dump()
    deserialized_row = GrahamStepsTableRowPydanticAdapter(**serialized_row)
    assert deserialized_row.regular_object() == table_row_adapter.regular_object()


def test_steps_table_adapter(table_adapter):
    regular_object = GrahamStepsTable(
        [Point(1, 1)],
        [GrahamStepsTableRow((Point(1, 1), Point(2, 2), Point(3, 3)), True)]
    )

    assert table_adapter.regular_object() == regular_object
    assert GrahamStepsTablePydanticAdapter.from_regular_object(regular_object) == table_adapter


def test_steps_table_adapter_serialization(table_adapter):
    serialized_table = table_adapter.model_dump()
    deserialized_table = GrahamStepsTablePydanticAdapter(**serialized_table)
    assert deserialized_table.regular_object() == table_adapter.regular_object()


from __future__ import annotations
from typing import Any, ClassVar, Optional, Iterable, Generator, Union
from enum import Enum
from copy import deepcopy
from typing_extensions import Literal
from pydantic import BaseModel
from PyCompGeomAlgorithms.core import PyCGAObject, Point, BinTreeNode, BinTree, ThreadedBinTreeNode, ThreadedBinTree
from PyCompGeomAlgorithms.dynamic_hull import DynamicHullNode, DynamicHullTree, SubhullNode, SubhullThreadedBinTree
from PyCompGeomAlgorithms.graham import GrahamStepsTableRow, GrahamStepsTable
from PyCompGeomAlgorithms.preparata import PreparataNode, PreparataThreadedBinTree
from PyCompGeomAlgorithms.quickhull import QuickhullNode, QuickhullTree


class SerializablePydanticModelWithPydanticFields(BaseModel):
    """
        A class to ensure correct serialization of Pydantic models that have other Pydantic models' instances as fields
        with possible cyclic references whose custom serialization is specified in those models.
    """
    def model_dump(self, *args, **kwargs):
        tmp_fields_dict = {}
        this_utility_class = SerializablePydanticModelWithPydanticFields

        for field, value in self.__dict__.items():
            tmp_fields_dict[field] = value
            
            if isinstance(value, this_utility_class):
                setattr(self, field, value.model_dump())
            elif isinstance(value, dict):
                setattr(
                    self,
                    field,
                    {
                        (k.model_dump() if isinstance(k, this_utility_class) else k): (v.model_dump() if isinstance(v, this_utility_class) else v)
                        for k, v in value.items()
                    }
                )
            elif not isinstance(value, str) and not isinstance(value, BaseModel) and isinstance(value, Iterable): # BaseModel's are Iterables, but we don't need them to be checked here 
                generator = (item.model_dump() if isinstance(item, this_utility_class) else item for item in value)
                setattr(self, field, generator if isinstance(value, Generator) else value.__class__(generator))
                
        result = super().model_dump(*args, **kwargs)

        for field, value in tmp_fields_dict.items():
            setattr(self, field, value)
        
        return result


class PydanticAdapter(SerializablePydanticModelWithPydanticFields):
    regular_class: ClassVar[type] = object

    @classmethod
    def from_regular_object(cls, obj, **kwargs):
        raise NotImplementedError

    def regular_object(self):
        return self.regular_class(**{
            field: self._regular_object(value)
            for field, value in self.__dict__.items()
        })
    
    @classmethod
    def _regular_object(cls, obj):
        if isinstance(obj, PydanticAdapter):
            return obj.regular_object()
        if isinstance(obj, dict):
            return {cls._regular_object(key): cls._regular_object(value) for key, value in obj.items()}
        if isinstance(obj, Iterable) and not isinstance(obj, str):
            generator = (cls._regular_object(item) for item in obj)
            return generator if isinstance(obj, Generator) else obj.__class__(generator)
        
        return obj

    def __eq__(self, other):
        return self.regular_object() == (other.regular_object() if isinstance(other, self.__class__) else other)
    
    def __hash__(self):
        return hash(self.regular_object())


class PointPydanticAdapter(PydanticAdapter):
    regular_class: ClassVar[type] = Point
    coords: tuple[float, ...]

    @classmethod
    def from_regular_object(cls, obj: Point, **kwargs):
        return cls(coords=obj.coords, **kwargs)

    def regular_object(self):
        return self.regular_class(*self.coords)


class BinTreeNodePydanticAdapter(PydanticAdapter):
    regular_class: ClassVar[type] = BinTreeNode
    data: Any # WARNING: do not leave Any in derived classes, override with specifying that type! (Otherwise it might come out as underserialized portion of JSON when deserializing)
    left: Optional[BinTreeNodePydanticAdapter] = None
    right: Optional[BinTreeNodePydanticAdapter] = None

    @classmethod
    def from_regular_object(cls, obj: BinTreeNode, **kwargs):
        return cls(
            data=pycga_to_pydantic(obj.data),
            left=pycga_to_pydantic(obj.left),
            right=pycga_to_pydantic(obj.right),
            **kwargs
        )
    
    def traverse_inorder(self, node=None, nodes=None):
        if node is None:
            node = self
        if nodes is None:
            nodes = []
        
        if node.left:
            self.traverse_inorder(node.left, nodes)
        
        nodes.append(node)

        if node.right:
            self.traverse_inorder(node.right, nodes)
        
        return nodes


class BinTreePydanticAdapter(PydanticAdapter):
    regular_class: ClassVar[type] = BinTree
    root: BinTreeNodePydanticAdapter

    @classmethod
    def from_regular_object(cls, obj: BinTree, **kwargs):
        return cls(root=pycga_to_pydantic(obj.root), **kwargs)
    
    def traverse_inorder(self):
        return self.root.traverse_inorder() if self.root else []


def serialize_threaded_bin_tree_root_or_tree(root_or_tree: ThreadedBinTreeNodePydanticAdapter | ThreadedBinTreePydanticAdapter, *args, **kwargs):
    """
        Serializes both a threaded bin tree or its root.
    """
    copy = deepcopy(root_or_tree)
    nodes_inorder = copy.traverse_inorder() if isinstance(copy, ThreadedBinTreeNodePydanticAdapter) else copy.root.traverse_inorder()
    is_circular = nodes_inorder and nodes_inorder[0].prev is nodes_inorder[-1]

    # Store node positions in inorder traversal instead of nodes to avoid infinite loops and be able to convert the tree to JSON
    for i, node in enumerate(nodes_inorder):
        node.prev = (i - 1) % len(nodes_inorder)
        node.next = (i + 1) % len(nodes_inorder)
    
    if not is_circular and nodes_inorder:
        nodes_inorder[0].prev = None
        nodes_inorder[-1].next = None

    return copy.model_dump(*args, **(kwargs | {'can_serialize': True}))


def deserialize_threaded_bin_tree_root(root):
    nodes_inorder = root.traverse_inorder()
    is_circular = (
        nodes_inorder
        and nodes_inorder[0].prev is not None
        and nodes_inorder[-1].next is not None
        and (
            (nodes_inorder[-1].next == 0)                     # if the object comes from JSON, this checks the node indices (see serialize_threaded_bin_tree_root)
            if isinstance(nodes_inorder[0].prev, int)
            else (nodes_inorder[0].prev is nodes_inorder[-1]) # if the object comes from Python objects, this checks the nodes 
        )
    )

    for i, node in enumerate(nodes_inorder):
        node.prev = nodes_inorder[i-1]
        node.next = nodes_inorder[(i + 1) % len(nodes_inorder)]
    
    if not is_circular and nodes_inorder:
        nodes_inorder[0].prev = None
        nodes_inorder[-1].next = None


class ThreadedBinTreeNodePydanticAdapter(BinTreeNodePydanticAdapter):
    regular_class: ClassVar[type] = ThreadedBinTreeNode
    left: Optional[ThreadedBinTreeNodePydanticAdapter] = None
    right: Optional[ThreadedBinTreeNodePydanticAdapter] = None
    prev: Optional[Union[ThreadedBinTreeNodePydanticAdapter, int]] = None
    next: Optional[Union[ThreadedBinTreeNodePydanticAdapter, int]] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        deserialize_threaded_bin_tree_root(self)  
    
    def model_dump(self, *args, **kwargs):
        if kwargs.get('can_serialize', False):
            kwargs.pop('can_serialize')
            return BaseModel.model_dump(self, *args, **kwargs)
        
        return serialize_threaded_bin_tree_root_or_tree(self, *args, **kwargs)

    def regular_object(self):
        return self.regular_class(
            self.data.regular_object() if isinstance(self.data, PydanticAdapter) else self.data,
            self.left.regular_object() if self.left else None,
            self.right.regular_object() if self.right else None
        )
    
    @classmethod
    def from_regular_object(cls, obj: ThreadedBinTreeNode, **kwargs):
        return super().from_regular_object(obj, prev=None, next=None, **kwargs)
        

class ThreadedBinTreePydanticAdapter(BinTreePydanticAdapter):
    regular_class: ClassVar[type] = ThreadedBinTree
    root: ThreadedBinTreeNodePydanticAdapter

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        deserialize_threaded_bin_tree_root(self.root)

    def model_dump(self, *args, **kwargs):
        if kwargs.get('can_serialize', False):
            kwargs.pop('can_serialize')
            return BaseModel.model_dump(self, *args, **kwargs)
        
        return serialize_threaded_bin_tree_root_or_tree(self, *args, **kwargs)

    @classmethod
    def from_regular_object(cls, obj: ThreadedBinTree, **kwargs):
        root = obj.root
        is_circular = root.leftmost_node.prev is root.rightmost_node or root.rightmost_node.next is root.leftmost_node
        root = pycga_to_pydantic(obj.root)
        nodes = root.traverse_inorder()

        for i, node in enumerate(nodes):
            node.prev = node.left if node.left else nodes[i-1]
            node.next = node.right if node.right else nodes[(i+1)%len(nodes)]
        
        if not is_circular and nodes:
            nodes[0].prev = None
            nodes[-1].next = None
        
        return cls(root=root, **kwargs)

    def regular_object(self):
        node = self.root
        while node.next is not None and node.next is not self.root:
            node = node.next
        
        is_circular = node.next is self.root
        regular_root = self.root.regular_object()

        return self.regular_class.from_iterable([node.data for node in regular_root.traverse_inorder()], is_circular)


def pydantic_to_pycga(obj: Any | PydanticAdapter | Iterable):    
    if isinstance(obj, PydanticAdapter):
        return obj.regular_object()

    if isinstance(obj, dict):
        return {pydantic_to_pycga(key): pydantic_to_pycga(value) for key, value in obj.items()}
    
    if isinstance(obj, Iterable) and not isinstance(obj, str):
        generator = (pydantic_to_pycga(obj) for obj in obj)
        return generator if isinstance(obj, Generator) else obj.__class__(generator)
    
    return obj


def pycga_to_pydantic(obj: Any | type | PyCGAObject | Iterable):
    if isinstance(obj, type):
        try:
            from .dynamic_hull import DynamicHullNodePydanticAdapter, DynamicHullTreePydanticAdapter, SubhullNodePydanticAdapter, SubhullThreadedBinTreePydanticAdapter
            from .graham import GrahamStepsTableRowPydanticAdapter, GrahamStepsTablePydanticAdapter
            from .preparata import PreparataNodePydanticAdapter, PreparataThreadedBinTreePydanticAdapter
            from .quickhull import QuickhullNodePydanticAdapter, QuickhullTreePydanticAdapter
            
            return {
                Point: PointPydanticAdapter,
                BinTreeNode: BinTreeNodePydanticAdapter,
                BinTree: BinTreePydanticAdapter,
                ThreadedBinTreeNode: ThreadedBinTreeNodePydanticAdapter,
                ThreadedBinTree: ThreadedBinTreePydanticAdapter,
                DynamicHullNode: DynamicHullNodePydanticAdapter,
                DynamicHullTree: DynamicHullTreePydanticAdapter,
                SubhullNode: SubhullNodePydanticAdapter,
                SubhullThreadedBinTree: SubhullThreadedBinTreePydanticAdapter,
                GrahamStepsTableRow: GrahamStepsTableRowPydanticAdapter,
                GrahamStepsTable: GrahamStepsTablePydanticAdapter,
                PreparataNode: PreparataNodePydanticAdapter,
                PreparataThreadedBinTree: PreparataThreadedBinTreePydanticAdapter,
                QuickhullNode: QuickhullNodePydanticAdapter,
                QuickhullTree: QuickhullTreePydanticAdapter,
            }[obj]
        except KeyError as e:
            raise KeyError("unknown PyCGA type") from e

    if isinstance(obj, PyCGAObject) and not isinstance(obj, Enum):
        return pycga_to_pydantic(obj.__class__).from_regular_object(obj)

    if isinstance(obj, dict):
        return {pycga_to_pydantic(key): pycga_to_pydantic(value) for key, value in obj.items()}
    
    if isinstance(obj, Iterable) and not isinstance(obj, str):
        generator = (pycga_to_pydantic(item) for item in obj)
        return generator if isinstance(obj, Generator) else obj.__class__(generator)
    
    return obj


from typing import ClassVar
from PyCompGeomAlgorithms.graham import graham, GrahamStepsTableRow, GrahamStepsTable
from .adapters import pycga_to_pydantic, pydantic_to_pycga, PydanticAdapter, PointPydanticAdapter
from .core import Task, Grader, Mistake, Answers
from .parsers import PointListGivenJSONParser


class GrahamGrader(Grader):
    @classmethod
    def grade_methods(cls):
        return [
            cls.grade_default,
            cls.grade_iterable,
            cls.grade_default,
            cls.grade_iterable,
            cls.grade_iterable,
            cls.grade_angles_less_than_pi,
            cls.grade_angles_greater_than_or_equal_to_pi,
            cls.grade_finalization
        ]
    
    @classmethod
    def grade_angles_less_than_pi(cls, answer, correct_answer, scorings):
        return [
            Mistake(scorings)
            for row, next_row in zip(answer, answer[1:])
            if row.is_angle_less_than_pi and (
                row.point_triple[1] != next_row.point_triple[0] or
                row.point_triple[2] != next_row.point_triple[1] or
                next_row.point_triple[2] != cls._next_point(answer.ordered_points, row.point_triple[2])
            )
        ]
    
    @classmethod
    def grade_angles_greater_than_or_equal_to_pi(cls, answer, correct_answer, scorings):
        return [
            Mistake(scorings)
            for row, next_row in zip(answer, answer[1:])
            if not row.is_angle_less_than_pi and (
                (
                    row.point_triple[0] != next_row.point_triple[0] or
                    row.point_triple[2] != next_row.point_triple[1] or
                    next_row.point_triple[2] != cls._next_point(answer.ordered_points, next_row.point_triple[1])
                ) if row.point_triple[0] == answer.ordered_points[0] else
                (
                    row.point_triple[0] != next_row.point_triple[1] or
                    row.point_triple[2] != next_row.point_triple[2] or
                    next_row.point_triple[0] != cls._prev_point(answer.rows, next_row)
                )
            )
        ]

    @classmethod
    def grade_finalization(cls, answer, correct_answer, scorings):
        return [
            Mistake(scorings)
            for row in answer
            if row.point_triple[1] == answer.ordered_points[0]
        ]
    
    @staticmethod
    def _prev_point(rows, row):
        i = rows.index(row)

        try:
            return next(r for r in reversed(rows[:i]) if r.point_triple[1] == row.point_triple[1]).point_triple[0]
        except StopIteration:
            return None
    
    @staticmethod
    def _next_point(ordered_points, point):
        try:
            return ordered_points[(ordered_points.index(point)+1) % len(ordered_points)]
        except (IndexError, ValueError):
            return None


class GrahamStepsTableRowPydanticAdapter(PydanticAdapter):
    regular_class: ClassVar[type] = GrahamStepsTableRow
    point_triple: tuple[PointPydanticAdapter, PointPydanticAdapter, PointPydanticAdapter]
    is_angle_less_than_pi: bool

    @classmethod
    def from_regular_object(cls, obj: GrahamStepsTableRow, **kwargs):
        return cls(
            point_triple=pycga_to_pydantic(obj.point_triple),
            is_angle_less_than_pi=obj.is_angle_less_than_pi,
            **kwargs
        )


class GrahamStepsTablePydanticAdapter(PydanticAdapter):
    regular_class: ClassVar[type] = GrahamStepsTable
    ordered_points: list[PointPydanticAdapter]
    rows: list[GrahamStepsTableRowPydanticAdapter]

    @classmethod
    def from_regular_object(cls, obj: GrahamStepsTable, **kwargs):
        return cls(
            ordered_points=pycga_to_pydantic(obj.ordered_points),
            rows=pycga_to_pydantic(obj.rows),
            **kwargs
        )


class GrahamAnswers(Answers):
    centroid: PointPydanticAdapter
    ordered_points: list[PointPydanticAdapter]
    origin: PointPydanticAdapter
    steps_table: GrahamStepsTablePydanticAdapter
    point_triples_: list[tuple[PointPydanticAdapter, PointPydanticAdapter, PointPydanticAdapter]] | None = None
    are_angles_less_than_pi_: list[bool] | None = None

    @classmethod
    def from_iterable(cls, iterable):
        centroid, ordered_points, origin, point_triples, are_angles_less_than_pi, steps_table, *rest = iterable        
        return cls(centroid=centroid, ordered_points=ordered_points, origin=origin, steps_table=steps_table, point_triples_=point_triples, are_angles_less_than_pi_=are_angles_less_than_pi)

    @property
    def point_triples(self):
        return self.point_triples_ or [row.point_triple for row in self.steps_table.rows]

    @property
    def are_angles_less_than_pi(self):
        return self.are_angles_less_than_pi_ or [row.is_angle_less_than_pi for row in self.steps_table.rows]

    def to_pydantic_list(self):
        return [
            self.centroid, self.ordered_points, self.origin, self.point_triples,
            self.are_angles_less_than_pi, self.steps_table, self.steps_table, self.steps_table
        ]
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            tmp_triples = other.point_triples_
            tmp_angles = other.are_angles_less_than_pi_

            other.point_triples_ = self.point_triples_
            other.are_angles_less_than_pi_ = self.are_angles_less_than_pi_

            are_equal = super().__eq__(other)

            other.point_triples_ = tmp_triples
            other.are_angles_less_than_pi_ = tmp_angles

            return are_equal

        return False


class GrahamTask(Task):
    algorithm = graham
    grader_class = GrahamGrader
    answers_class = GrahamAnswers
    given_parser_class = PointListGivenJSONParser



from math import isclose
from copy import deepcopy
from PyCompGeomAlgorithms.core import Point
from PyCompGeomAlgorithms.preparata import PreparataThreadedBinTree
from AlgoGrade.preparata import PreparataGrader, PreparataTask, PreparataAnswers
from AlgoGrade.adapters import pycga_to_pydantic
from AlgoGrade.core import Scoring


points = [Point(1, 1), Point(1, 5), Point(5, 3), Point(1, 11), Point(6, 1), Point(10, 1)]
givens = (points,)
task_class = PreparataTask
scorings = [
    Scoring(max_grade=0.25, fine=0.25),
    Scoring(max_grade=0.25, fine=0.25, repeat_fine=1.5),
    Scoring(max_grade=0.25, fine=0.25, repeat_fine=1),
    Scoring(max_grade=0.25, fine=0.25, repeat_fine=1.5)
]
correct_pycga_answers = task_class.solve_as_pycga_list(givens)
correct_pydantic_answers = task_class.solve_as_pydantic_list(givens)
correct_answers_wrapper = task_class.solve_as_answers_wrapper(givens)


def test_preparata_all_correct():
    hull0 = [Point(1, 1), Point(1, 5), Point(5, 3)]
    hull = [Point(1, 1), Point(1, 11), Point(10, 1)]
    tree0 = PreparataThreadedBinTree.from_iterable(hull0)
    left_paths = [
        [Point(1, 5), Point(5, 3)],
        [Point(1, 11), Point(5, 3), Point(1, 1)],
        [Point(1, 11), Point(6, 1), Point(1, 1)]
    ]
    right_paths = [
        [Point(1, 5), Point(1, 1)],
        [Point(1, 11)],
        [Point(1, 11)]
    ]
    deleted_points = [[Point(1, 5)], [Point(5, 3)], [Point(6, 1)]]
    hulls = [
        [Point(1, 1), Point(1, 11), Point(5, 3)],
        [Point(1, 1), Point(1, 11), Point(6, 1)],
        hull
    ]
    trees = [PreparataThreadedBinTree.from_iterable(hulls[0]), PreparataThreadedBinTree.from_iterable(hulls[1])]

    pycga_answers = [(hull0, tree0), (left_paths, right_paths), deleted_points, (hulls, trees)]
    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 1)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 1)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 1)


def test_preparata_grader_incorrect_first_hull():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    first_hull = pycga_answers[0][0]
    first_hull[0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0.75)


def test_preparata_grader_incorrect_first_tree():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    first_tree = pycga_answers[0][1]
    first_tree.root.data = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0.75)


def test_preparata_grader_incorrect_left_paths_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    left_paths = pycga_answers[1][0]
    left_paths[0][0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0.75)


def test_preparata_grader_incorrect_left_paths_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    left_paths = pycga_answers[1][0]
    left_paths[0][0] = Point(100, 100)
    left_paths[0][1] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, -0.5)


def test_preparata_grader_incorrect_right_paths_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    right_paths = pycga_answers[1][1]
    right_paths[0][0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0.75)


def test_preparata_grader_incorrect_right_paths_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    right_paths = pycga_answers[1][1]
    right_paths[0][0] = Point(100, 100)
    right_paths[0][1] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, -0.5)


def test_preparata_grader_incorrect_left_and_right_paths():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    left_paths = pycga_answers[1][0]
    right_paths = pycga_answers[1][1]
    left_paths[0][0] = Point(100, 100)
    right_paths[0][0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, -0.5)


def test_preparata_grader_incorrect_deleted_points_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    deleted_points = pycga_answers[2]
    deleted_points[0][0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0.75)


def test_preparata_grader_incorrect_deleted_points_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    deleted_points = pycga_answers[2]
    deleted_points[0][0] = Point(100, 100)
    deleted_points[1][0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0)


def test_preparata_grader_incorrect_hulls_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    hulls = pycga_answers[3][0]
    hulls[0][0] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0.75)


def test_preparata_grader_incorrect_hulls_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    hulls = pycga_answers[3][0]
    hulls[0][0] = Point(100, 100)
    hulls[1][1] = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, -0.5)


def test_preparata_grader_incorrect_trees_single():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    trees = pycga_answers[3][1]
    trees[0].root.data = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, 0.75)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, 0.75)


def test_preparata_grader_incorrect_trees_repeated():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    trees = pycga_answers[3][1]
    trees[0].root.data = Point(100, 100)
    trees[0].root.left.data = Point(-1, -1)
    trees[1].root.data = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, -0.5)


def test_preparata_grader_incorrect_trees_and_hulls():
    pycga_answers = deepcopy(correct_pycga_answers)
    
    hulls = pycga_answers[3][0]
    hulls[0][0] = Point(100, 100)
    hulls[1][1] = Point(100, 100)
    
    trees = pycga_answers[3][1]
    trees[0].root.data = Point(100, 100)
    trees[0].root.left.data = Point(-1, -1)
    trees[1].root.data = Point(100, 100)

    pydantic_answers = pycga_to_pydantic(pycga_answers)
    answers_wrapper = PreparataAnswers.from_iterable(pydantic_answers)

    total_grade, answer_grades = PreparataGrader.grade_pycga(pycga_answers, correct_pycga_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_pydantic(pydantic_answers, correct_pydantic_answers, scorings)
    assert isclose(total_grade, -0.5)

    total_grade, answer_grades = PreparataGrader.grade_answers_wrapper(answers_wrapper, correct_answers_wrapper, scorings)
    assert isclose(total_grade, -0.5)



from pytest import fixture
from PyCompGeomAlgorithms.core import Point, BinTree
from PyCompGeomAlgorithms.quickhull import QuickhullNode
from AlgoGrade.adapters import PointPydanticAdapter
from AlgoGrade.quickhull import QuickhullNodePydanticAdapter, QuickhullTreePydanticAdapter


@fixture
def regular_node():
    return QuickhullNode([Point(1, 1), Point(2, 2)], subhull=[Point(1, 1), Point(2, 2)])


@fixture
def adapter_node():
    return QuickhullNodePydanticAdapter(
        data=[
            PointPydanticAdapter(coords=(1, 1)),
            PointPydanticAdapter(coords=(2, 2))
        ],
        subhull=[
            PointPydanticAdapter(coords=(1, 1)),
            PointPydanticAdapter(coords=(2, 2))
        ]
    )


def test_quickhull_node_adapter(regular_node, adapter_node):
    assert adapter_node.regular_object() == regular_node
    assert QuickhullNodePydanticAdapter.from_regular_object(regular_node) == adapter_node


def test_quickhull_node_adapter_serialization(adapter_node):
    serialized_node = adapter_node.model_dump()
    deserialized_node = QuickhullNodePydanticAdapter(**serialized_node)
    assert deserialized_node.regular_object() == adapter_node.regular_object()


def test_quickhull_tree_adapter(regular_node, adapter_node):
    regular_object = BinTree(regular_node)
    adapter_object = QuickhullTreePydanticAdapter(root=adapter_node)

    assert adapter_object.regular_object() == regular_object
    assert QuickhullTreePydanticAdapter.from_regular_object(regular_object) == adapter_object


def test_quickhull_tree_adapter_serialization(adapter_node):
    adapter_tree = QuickhullTreePydanticAdapter(root=adapter_node)
    serialized_tree = adapter_tree.model_dump()
    deserialized_tree = QuickhullTreePydanticAdapter(**serialized_tree)
    assert deserialized_tree.regular_object() == adapter_tree.regular_object()


from copy import deepcopy
from pytest import fixture
from PyCompGeomAlgorithms.core import Point, BinTreeNode, BinTree, ThreadedBinTreeNode, ThreadedBinTree
from AlgoGrade.adapters import PointPydanticAdapter, BinTreeNodePydanticAdapter, BinTreePydanticAdapter, ThreadedBinTreeNodePydanticAdapter, ThreadedBinTreePydanticAdapter


@fixture
def coords():
    return 1, 2, 3


@fixture
def regular_point(coords):
    return Point(*coords)


@fixture
def adapter_point(coords):
    return PointPydanticAdapter(coords=coords)


def test_point_adapter(regular_point, adapter_point):
    assert adapter_point.regular_object() == regular_point


def test_point_adapter_serialization(adapter_point):
    serialized_point = adapter_point.model_dump()
    deserialized_point = PointPydanticAdapter(**serialized_point)
    assert deserialized_point.regular_object() == adapter_point.regular_object()


@fixture
def adapter_root():
    adapter_root = BinTreeNodePydanticAdapter(data=1)
    adapter_left = BinTreeNodePydanticAdapter(data=2)
    adapter_right = BinTreeNodePydanticAdapter(data=3)
    adapter_root.left = adapter_left
    adapter_root.right = adapter_right

    return adapter_root


@fixture
def regular_root():
    return BinTreeNode(1, left=BinTreeNode(2), right=BinTreeNode(3))


def test_bin_tree_node_adapter(adapter_root, regular_root):
    assert adapter_root.regular_object() == regular_root


def test_bin_tree_node_adapter_serialization(adapter_root):
    serialized_root = adapter_root.model_dump()
    deserialized_root = BinTreeNodePydanticAdapter(**serialized_root)
    assert deserialized_root.regular_object() == adapter_root.regular_object()


def test_bin_tree_adapter(adapter_root, regular_root):
    adapter_tree = BinTreePydanticAdapter(root=adapter_root)
    regular_tree = BinTree(regular_root)
    assert adapter_tree.regular_object() == regular_tree


def test_bin_tree_adapter_serialization(adapter_root):
    adapter_tree = BinTreePydanticAdapter(root=adapter_root)
    serialized_tree = adapter_tree.model_dump()
    deserialized_tree = BinTreePydanticAdapter(**serialized_tree)
    assert deserialized_tree.regular_object() == adapter_tree.regular_object()


@fixture
def adapter_tbt_root_circular():
    adapter_root = ThreadedBinTreeNodePydanticAdapter(data=1)
    adapter_left = ThreadedBinTreeNodePydanticAdapter(data=2)
    adapter_right =ThreadedBinTreeNodePydanticAdapter(data=3)
    adapter_root.left = adapter_left
    adapter_root.right = adapter_right

    adapter_root.prev = adapter_root.left
    adapter_root.next = adapter_root.right
    adapter_left.prev = adapter_right
    adapter_left.next = adapter_root
    adapter_right.prev = adapter_root
    adapter_right.next = adapter_left

    return adapter_root


@fixture
def regular_tbt_root_circular():
    left = ThreadedBinTreeNode(2)
    right = ThreadedBinTreeNode(3)
    root = ThreadedBinTreeNode(1, left, right)
    root.prev = left
    root.next = right
    left.prev = right
    left.next = root
    right.prev = root
    right.next = left

    return root


@fixture
def adapter_tbt_root(adapter_tbt_root_circular):
    adapter_tbt_root = deepcopy(adapter_tbt_root_circular)
    adapter_tbt_root.left.prev = None
    adapter_tbt_root.right.next = None

    return adapter_tbt_root


@fixture
def regular_tbt_root(regular_tbt_root_circular):
    regular_tbt_root = deepcopy(regular_tbt_root_circular)
    regular_tbt_root.left.prev = None
    regular_tbt_root.right.next = None

    return regular_tbt_root


def test_threaded_bin_tree_node_adapter(adapter_tbt_root, regular_tbt_root):
    assert adapter_tbt_root.regular_object() == regular_tbt_root


def test_threaded_bin_tree_node_adapter_circular(adapter_tbt_root_circular, regular_tbt_root_circular):
    assert adapter_tbt_root_circular.regular_object() == regular_tbt_root_circular


def test_threaded_bin_tree_node_adapter_seriaization(adapter_tbt_root):
    serialized_root = adapter_tbt_root.model_dump()
    deserialized_root = ThreadedBinTreeNodePydanticAdapter(**serialized_root)
    assert deserialized_root.regular_object() == adapter_tbt_root.regular_object()


def test_threaded_bin_tree_node_adapter_circular_seriaization(adapter_tbt_root_circular):
    serialized_root = adapter_tbt_root_circular.model_dump()
    deserialized_root = ThreadedBinTreeNodePydanticAdapter(**serialized_root)
    assert deserialized_root.regular_object() == adapter_tbt_root_circular.regular_object()


def test_threaded_bin_tree_adapter(adapter_tbt_root, regular_tbt_root):
    adapter_tbt = ThreadedBinTreePydanticAdapter(root=adapter_tbt_root)
    regular_tbt = ThreadedBinTree(regular_tbt_root)
    assert adapter_tbt.regular_object() == regular_tbt


def test_threaded_bin_tree_adapter_circular(adapter_tbt_root_circular, regular_tbt_root_circular):
    adapter_tbt = ThreadedBinTreePydanticAdapter(root=adapter_tbt_root_circular)
    regular_tbt = ThreadedBinTree(regular_tbt_root_circular)
    assert adapter_tbt.regular_object() == regular_tbt


def test_threaded_bin_tree_adapter_serialization(adapter_tbt_root):
    adapter_tbt = ThreadedBinTreePydanticAdapter(root=adapter_tbt_root)
    serialized_tbt = adapter_tbt.model_dump()
    deserialized_tbt = ThreadedBinTreePydanticAdapter(**serialized_tbt)

    assert deserialized_tbt.regular_object() == adapter_tbt.regular_object()


def test_threaded_bin_tree_adapter_circular_serialization(adapter_tbt_root_circular):
    adapter_tbt = ThreadedBinTreePydanticAdapter(root=adapter_tbt_root_circular)
    serialized_tbt = adapter_tbt.model_dump()
    deserialized_tbt = ThreadedBinTreePydanticAdapter(**serialized_tbt)

    assert deserialized_tbt.regular_object() == adapter_tbt.regular_object()